%{
#include "tcplogd.h"
#include "util.h"

/* lex stuff */
#define  yywrap() 1
#define YY_NO_UNPUT 1
#define YY_NO_TOP_STATE 1

/* iternal conditions */
#define LOGGING_COND 1
#define IGNOREPRTS_COND 2
#define FLOOD_COND 3
#define TRUSTED_COND 4

int linenum = 1;
unsigned char field=0;

int floodlimit=DEFAULT_FLOODLIMIT;
int floodtimeout=DEFAULT_FLOODTIMEOUT;
struct tcpflags *tcpflag_current=NULL,*tcpflag_start=NULL;

struct ports *ports_start,*ports_current;
struct hosts *hosts_start=NULL,*hosts_current=NULL,*hosts_temp;

char *tmp;
int  i_temp;
struct in_addr ia_tmp;
int *intptr=NULL;
int errors=0;
unsigned long int buf[HBUFSIZE*4+1]; /* for passing up IP addresses */
/* FLAGS */

unsigned char *tcpex[] = {
"TCP_FIN","TCP_SYN","TCP_RST","TCP_PSH","TCP_ACK","TCP_URG","NON-EXISTANT",
(unsigned char *)NULL};
unsigned char *tcpbt="\x1\x2\x4\x8\x10\x20\xff\x0";

unsigned char *facilex[]={
"EMERG","ALERT","CRIT","ERR","WARNING","NOTICE","INFO","DEBUG","NON-EXISTANT",
(unsigned char *)NULL};

void conf_error(int lnum, char *tok) {
	 fprintf(stderr,"%s:error at line %i: %s\n",TCPLOGD_CF,lnum,tok);
	 errors++;
}

int setflag(struct tcpflags *ptcp,char *tcpfl)
{


unsigned char *foo,**bar,*bts;

foo=tcpfl;
if (*foo=='!') foo++; /* we shall invert it then */

bar=tcpex;
bts=tcpbt;
while(strcmp(*bar++,foo) && *bar) bts++;

if(*tcpfl=='!') ptcp->flags&=(*bts^0xff);
	else ptcp->flags|=*bts;
	
ptcp->flags&=0x3f; /* clear reserved bits */
if (*bts==0xff)  conf_error(linenum,"Non-existant facility");
return 0;
}

int setfacility(struct tcpflags *ptcp,char *facil)
{
unsigned char **bar,bts;


bar=facilex;
bts=0;
while(strcmp(*bar++,facil) && *bar) bts++;

ptcp->facility=bts;
if (bts>7)  conf_error(linenum,"Non-existant facility");
return 0;
}

int setname(struct tcpflags *ptcp,char *name)
{
snprintf(ptcp->name,BUFSIZE,"%s",name);
return 0;
}
%}

%option stack

%x LOGGING IGNOREPORTS TRUSTED FLOOD LCOMMENT OP RIGHT

eol	(\r)|(\n)|(\r\n)
openp  \{
closep  \}
TCP_FLAG (TCP_URG)|(TCP_ACK)|(TCP_PSH)|(TCP_RST)|(TCP_SYN)|(TCP_FIN)|(!TCP_URG)|(!TCP_ACK)|(!TCP_PSH)|(!TCP_RST)|(!TCP_SYN)|(!TCP_FIN)
LOG_FACILITY (EMERG)|(ALERT)|(CRIT)|(ERR)|(WARNING)|(NOTICE)|(INFO)|(DEBUG)
PROBENAME \"[^"]*\"
OPS	[=&]
COMMENT (#.*((\r)|(\n)|(\r\n)))|(\/\/.*((\r)|(\n)|(\r\n)))
BOOL	(yes)|(no)
WHITESPACE [ \t]+
DIGIT	   [0-9]
IP	  DIGIT+.DIGIT+.DIGIT+.DIGIT+
HOSTNAME  [a-z0-9.]+
NONE	   [ \t\r\n]
%%

<*>{COMMENT} /* comment */
"/*"		 yy_push_state(LCOMMENT); /* long comment */
<LCOMMENT>[^*\n]*        /* eat anything that's not a '*' */
<LCOMMENT>"*"+[^*/\n]*   /* eat up '*'s not followed by '/'s */
<LCOMMENT>"*"+"/" yy_pop_state();

					    

<*>{WHITESPACE}
<LOGGING>{eol}		{
		if(tcpflag_current->facility!=-1) { /* if previous used */
		tmp=(char *)malloc(sizeof(struct tcpflags));
		if(!tmp) { conf_error(linenum,"Can not allocate memory");
			   yyterminate();
			   }
		tcpflag_current->next=(struct tcpflags *)tmp;
		tcpflag_current=(struct tcpflags *)tmp;
		bzero(tcpflag_current,sizeof(struct tcpflags));
		tcpflag_current->facility=-1;
		}
		}
<*>{eol}       { linenum++; }
"logging"{NONE}*/{openp}     { field=LOGGING_COND;
			if(tcpflag_start!=NULL) {
			conf_error(linenum,"logging appears more than once");
			yyterminate();
			}
			tcpflag_current=
			(struct tcpflags *)malloc(sizeof(struct tcpflags));
		if(!tcpflag_current) { 
			conf_error(linenum,"Can not allocate memory");
			   yyterminate();
			   }
			tcpflag_start=tcpflag_current;
			bzero(tcpflag_current,sizeof(struct tcpflags));
			tcpflag_current->facility=-1;
		}
"ignore_ports"{NONE}*/{openp}	{ field=IGNOREPRTS_COND;
				ports_start=
				(struct ports *)malloc(sizeof(struct ports));
		if(!ports_start) { 
			conf_error(linenum,"Can not allocate memory");
			   yyterminate();
			   }
				ports_current=ports_start;
				bzero(ports_start,sizeof(struct ports));
				ports_start->port=-1;
		}
"flood"{NONE}*/{openp}		{ field=FLOOD_COND;
		}

"trusted"{NONE}*/{openp}	{ field=TRUSTED_COND;
				hosts_start=
				(struct hosts *)malloc(sizeof(struct hosts));
				if(!hosts_start) {
				conf_error(linenum,"Can not allocate memory");
			   	yyterminate();
			   	}
				hosts_current=hosts_start;
				hosts_start->prev=NULL;
				hosts_start->next=NULL;
				
		}

{openp}		{ switch(field) {
			case LOGGING_COND:	yy_push_state(LOGGING);
						
						break;
			case IGNOREPRTS_COND:	yy_push_state(IGNOREPORTS);
						break;
			case FLOOD_COND:	yy_push_state(FLOOD);
						break;
			case TRUSTED_COND:	yy_push_state(TRUSTED);
						break;
			default:	conf_error(linenum,yytext);
					yyterminate();
				}/* switch */
		}
			
<LOGGING,IGNOREPORTS,FLOOD,TRUSTED>{closep}       { yy_pop_state(); }

<LOGGING>{TCP_FLAG}/{OPS} { setflag(tcpflag_current,yytext);
				}
<LOGGING>{OPS}/{TCP_FLAG}   /* we are expecting just next tcp flag */ 
<LOGGING>{OPS}/{LOG_FACILITY} /* or facility */
<LOGGING>{LOG_FACILITY}/{NONE}*{PROBENAME}  {
			setfacility(tcpflag_current,yytext);
			}
<LOGGING>{PROBENAME}	{
			setname(tcpflag_current,yytext); }
<IGNOREPORTS>{DIGIT}+-{DIGIT}+ {
			char *foo;
			foo=strchr(yytext,'-');
			if (!foo) { conf_error(linenum,yytext);
				   }

			if (atoi(yytext)<atoi(foo+1)) { 
			ports_current->port=atoi(yytext);
			ports_current->hport=atoi(foo+1);
			} else {
			/* some idiot may swap higher and lower ports */
			ports_current->hport=atoi(yytext);
			ports_current->port=atoi(foo+1);
			}
			ports_current->range=0xff;
			if (atoi(yytext)<0 || atoi(yytext)>65535 ||
			    atoi(foo+1)<0 || atoi(foo+1)>65535)  
			    conf_error(linenum,"port number(s) is out of range. (0-65535)\n");
			
			tmp=(char *)malloc(sizeof(struct ports));
		if(!tmp) { conf_error(linenum,"Can not allocate memory");
			   yyterminate();
			   }
			ports_current->next=(struct ports *)tmp;
			ports_current=(struct ports *)tmp;
			bzero(ports_current,sizeof(struct ports));
			ports_current->port=-1;
			
	
			}

<IGNOREPORTS>{DIGIT}+- {
			ports_current->port=atoi(yytext);
			ports_current->hport=65535;
			ports_current->range=0xff;

			if (atoi(yytext)<0 || atoi(yytext)>65535)
			 conf_error(linenum,"port number(s) is out of range. (0-65535)\n");

			tmp=(char *)malloc(sizeof(struct ports));
		if(!tmp) { conf_error(linenum,"Can not allocate memory");
			   yyterminate();
			   }
			ports_current->next=(struct ports *)tmp;
			ports_current=(struct ports *)tmp;
			bzero(ports_current,sizeof(struct ports));
			ports_current->port=-1;
			
	
			}
<IGNOREPORTS>-{DIGIT}+ {
			ports_current->port=0;
			ports_current->hport=atoi(yytext+1);
			ports_current->range=0xff;
			if (atoi(yytext)<0 || atoi(yytext)>65535)
			 conf_error(linenum,"port number(s) is out of range. (0-65535)\n");
			tmp=(char *)malloc(sizeof(struct ports));
		if(!tmp) { conf_error(linenum,"Can not allocate memory");
			   yyterminate();
			   }
			ports_current->next=(struct ports *)tmp;
			ports_current=(struct ports *)tmp;
			bzero(ports_current,sizeof(struct ports));
			ports_current->port=-1;
			
			}

<IGNOREPORTS>{DIGIT}+	{
			ports_current->port=atoi(yytext);
			ports_current->range=0;
			if (atoi(yytext)<0 || atoi(yytext)>65535)
			 conf_error(linenum,"port number(s) is out of range. (0-65535)\n");
			tmp=(char *)malloc(sizeof(struct ports));
		if(!tmp) { conf_error(linenum,"Can not allocate memory");
			   yyterminate();
			   }
			ports_current->next=(struct ports *)tmp;
			ports_current=(struct ports *)tmp;
			bzero(ports_current,sizeof(struct ports));
			ports_current->port=-1;
			}

<FLOOD>"limit"{NONE}*/[=]	{ intptr=&floodlimit;
			}
<FLOOD>"timeout"{NONE}*/[=]	{ intptr=&floodtimeout;
			}
<FLOOD>[=]		{ 
			}
<FLOOD>{DIGIT}+		{ *intptr=atoi(yytext);
			}

<TRUSTED>{HOSTNAME}	{
			i_temp=resolve(yytext,buf);
			if(i_temp) {
				i_temp=0;
					while(buf[i_temp]!=0) {
					hosts_current->ip=buf[i_temp];
					hosts_temp= (struct hosts *)
						malloc(sizeof(struct hosts));
			if(!hosts_temp) {
				conf_error(linenum,"Can not allocate memory");
			        yyterminate();
				}
					hosts_temp->prev=hosts_current;
					hosts_current->next=hosts_temp;
					hosts_temp->next=NULL;
					hosts_current=hosts_temp;
					i_temp++;
				}
				} else  /* never give up ;-) */
				printf("WARNING: host %s not found - ignored\n",yytext); 

			}

<*>.		{ conf_error(linenum,yytext);
		  yyterminate();
		}

%%

void show_conf(void) {
struct ports *pports=ports_start;
struct tcpflags *ptcp=tcpflag_start;
printf("ignoring ports:");
while(pports->port!=-1 && pports->next!=NULL){ 
	if(pports->range)
	printf(" %i-%i",pports->port,pports->hport);
	else
	printf(" %i",pports->port);
	pports=pports->next;
}
printf("\nlogging packets:\n");
while(ptcp->facility!=-1 && ptcp->next!=NULL) {
printf("type: %s\tmask: %0x\tfacility: %s\n",
ptcp->name, ptcp->flags,facilex[ptcp->facility]);
ptcp=ptcp->next;
}
printf("\n flood limit: %i flood timeout: %i\n",floodlimit,floodtimeout);
printf("\ntrusted hosts\n");
hosts_temp=hosts_start;
while(hosts_temp!=NULL && hosts_temp->next!=NULL) {
	printf("%s ", hostlookup(hosts_temp->ip));
	ia_tmp.s_addr=hosts_temp->ip;
	printf("[%s]\n",inet_ntoa(ia_tmp));
	hosts_temp=hosts_temp->next;
	}
		


}
int scan_conf(void) {

yyin=fopen(TCPLOGD_CF,"r");
if (!yyin) { fprintf(stderr,"can not open config file:%s\n",TCPLOGD_CF);
	     return -1;
	   }
yylex();
fclose(yyin);
if( errors!=0) return -1;
return 0;
}

/*
this would be used to check config file
*/
#ifdef CONFCHECK
void main(void) {
printf("\n Checking tcplogger config file [%s]\n\n",TCPLOGD_CF);

yyin=fopen(TCPLOGD_CF,"r");
if (!yyin) { fprintf(stderr,"can not open config file:%s\n",TCPLOGD_CF);
	     exit(1);
	   }
yylex();
printf("encounted %i errors. Scanned configuration is:\n",errors);
show_conf();
fclose(yyin);
}
#endif
